// OBGE Custom Water
// by Tomerk

/* User Editable Input Variables, Default Values In The Comments */

extern float causticsStrength = 0.35; 
//The strength of the caustics (0.35)

extern float waveWidth = 1.0;

extern float waveSpeed = 1.0;
extern float turbidity = 1.0; 
//Density of particles in the water. Lower values make water more transparent (1.0)

extern float inScattCoeff = 0.13; 
//Scatter Coefficient of the water (0.13)

extern float4 inExtCoeff = float4(0.3984, 0.2684, 0.2800, 1); 
//Extinction Coefficient of the water (0.3984, 0.2684, 0.2800, 1)

extern float depthDarkness = 0.1; 
//Affects how quickly the water volume gets darker with depth underwater (0.1)

/* Plugin Input Variables */

extern float waterHeight = 0; //Height of the water in the cell
extern float4 sunColor = float4(1, 0.71, 0.33, 1);//

extern float fogStart = 4096; //above water fog start
extern float fogEnd = 170000; //above water fog end
extern float4 fogColor = float4(0.39,0.57,0.75,1); //above water fog color

extern bool lava = 0; //Whether its currently rendering lava

/* End of Input Variables */

static const float3 extCoeff = inExtCoeff.rgb * turbidity;
static const float scattCoeff = inScattCoeff * turbidity;

texture lastpass;
texture Depth;
texture causticTexture
< 
        string filename="effects\\water_NRM_LOD.dds"; 
>;

sampler frameSampler = sampler_state { texture = <lastpass>; AddressU=Clamp; AddressV=Clamp;};
sampler depthSampler = sampler_state { texture = <Depth>; AddressU=Clamp; AddressV=Clamp; };
sampler causticSampler = sampler_state { texture = <causticTexture>; AddressU=Wrap; AddressV=Wrap; MagFilter=LINEAR; MinFilter = LINEAR; MipFilter = NONE;};

matrix m44world;
matrix m44view;
matrix m44proj;
float4 f4SunDir;
float2 rcpres;
float4 obge_Tick;

static const float animFrames = 32.0f;
static const float3 eyepos = float3( -m44world[3][0], -m44world[3][1], -m44world[3][2] - waterHeight );

static const float nearZ = m44proj._43 / m44proj._33;
static const float farZ = (m44proj._33 * nearZ) / (m44proj._33 - 1.0f);
static const float Zmul = nearZ * farZ;
static const float Zdiff = farZ - nearZ;
static const float depthRange = nearZ - farZ;
static const float frame = obge_Tick.y * waveSpeed / 1500;

static const matrix <float, 4, 4> ditherMat = { 0.0588, 0.5294, 0.1765, 0.6471,
												0.7647, 0.2941, 0.8824, 0.4118,
												0.2353, 0.7059, 0.1176, 0.5882,
												0.9412, 0.4706, 0.8235, 0.3259 };
							   
struct VSOUT
{
	float4 vertPos : POSITION;
	float2 UVCoord : TEXCOORD0;
};

struct VSIN
{
	float4 vertPos : POSITION0;
	float2 UVCoord : TEXCOORD0;
};

VSOUT FrameVS(VSIN IN)
{
	VSOUT OUT = (VSOUT)0.0f;	// initialize to zero, avoid complaints.
	OUT.vertPos = IN.vertPos;
	OUT.UVCoord = IN.UVCoord;
	return OUT;
}

float3 toWorld(float2 tex)
{
    float3 v = float3(m44view[0][2], m44view[1][2], m44view[2][2]);
    v += (1/m44proj[0][0] * (2*tex.x-1)).xxx * float3(m44view[0][0], m44view[1][0], m44view[2][0]);
    v += (-1/m44proj[1][1] * (2*tex.y-1)).xxx * float3(m44view[0][1], m44view[1][1], m44view[2][1]);
    return v;
}

float readDepth(in float2 coord : TEXCOORD0)
{
	float posZ = tex2D(depthSampler, coord).x;
	posZ = Zmul / ((posZ * Zdiff) - farZ);

	return posZ;
}

float2 get2DTex( float3 tex )
{
	float2 return_tex = float2(0, tex.y);
	return_tex.x = frac(tex.x)/animFrames + round( frac(tex.z)*(animFrames) )/animFrames;
	return return_tex;
}

float4 Water( VSOUT IN ) : COLOR0
{
    float4 color = tex2D(frameSampler, IN.UVCoord);
    float depth = readDepth(IN.UVCoord);
    float3 camera_vector = toWorld(IN.UVCoord);
	float3 norm_camera_vector = normalize( camera_vector );
    float3 world_pos = eyepos + camera_vector*depth;
	
	float eyeDist = eyepos.z/camera_vector.z;
	bool clipping = eyeDist > nearZ;
	
	/* Render water surface from above water */
    if ( (world_pos.z < 0) && (eyepos.z >= 0) && (clipping) )
    {
		float4 water_result = color;
		float uw_pos = world_pos.z / camera_vector.z;
		float2 surfPos = world_pos.xy - camera_vector.xy * uw_pos;
			
		//Render Caustics
		float3 dx = ddx(world_pos);
    	float3 dy = ddy(world_pos);
    	float3 waterfloorNorm = normalize(cross(dx,dy));
	
		float3 causticsPos = world_pos - f4SunDir.xyz * (world_pos.z / f4SunDir.z);
		float caustics = tex2Dlod(causticSampler, float4(get2DTex(float3(causticsPos.xy / (512*waveWidth),frame)),0,0) ).b;
		float causticsAngle = saturate( dot(-waterfloorNorm, f4SunDir.xyz) );
		water_result.rgb *=  0.7 + causticsStrength * caustics * causticsAngle * sunColor.xyz;
		
		//Calculate Refraction color
        water_result.rgb *= exp( -extCoeff * (uw_pos - world_pos.z) / 70 );
		
		float SinBoverSinA = -norm_camera_vector.z;
		float3 waterVolColor = scattCoeff * fogColor.xyz / ( extCoeff * (1 + SinBoverSinA) );
		
		//Lava Rendering
		if (lava)
			waterVolColor = float3(1,0.33,0.1);
			
		waterVolColor *= 1 - exp( -extCoeff * (1 + SinBoverSinA) * uw_pos / 70 );
		
		water_result.rgb += waterVolColor;

		float2 screenPos = IN.UVCoord / rcpres;
		water_result.rgb += ditherMat[ screenPos.x%4 ][ screenPos.y%4 ] / 255;

		color.rgb = water_result.rgb;
    }
	
	/* Underwater rendering */
	else if ( (eyepos.z < 0) )
	{
		/* Render Underwater Caustics */
		float sunDist = max( -world_pos.z / f4SunDir.z, 0 );
		if ( world_pos.z < 0 )
		{
			float3 dx = ddx(world_pos);
			float3 dy = ddy(world_pos);
			float3 waterfloorNorm = normalize(cross(dx,dy));
		
			float3 causticsPos = world_pos + f4SunDir.xyz * sunDist;
			float caustics = tex2Dlod(causticSampler, float4(get2DTex(float3(causticsPos.xy / (512*waveWidth),frame)),0,0) ).b;
			float causticsAngle = saturate( dot(-waterfloorNorm, f4SunDir.xyz) );
			color.rgb *=  0.7 + causticsStrength * caustics * causticsAngle * sunColor.xyz;
		}
		
		if (depth / depthRange < 0.999)
		{
			/* Render Above Water Fog */
			float fog = 0;
			if (!clipping)
				fog = saturate( (min(depth,abs(eyeDist))*length(camera_vector) - fogStart)/(fogEnd-fogStart) );
			else
				fog = saturate( (depth*length(camera_vector) - fogStart)/(fogEnd-fogStart) );
			
			color = lerp(color, fogColor, fog);
		}
		
		/* Render Underwater Light Scattering and Absorption */
		float waterfogDist = depth;
		float3 waterVolColor = float3(0,0,0);
		float SinBoverSinA = abs(norm_camera_vector.z);

		if (camera_vector.z > 0)
		{
			float SurfRay = -eyepos.z / camera_vector.z;
			waterfogDist = min( SurfRay, depth );
			
			float v = saturate(SinBoverSinA) * 0.5;
			waterVolColor = scattCoeff * fogColor.xyz / ( extCoeff * (1 - v) );
			
			if (lava)
				waterVolColor = float3(1,0.33,0.1);
				
			waterVolColor *= ( exp( extCoeff * (1 - v) * eyepos.z*depthDarkness / 70 ) - exp( -extCoeff * (1 - v) * (waterfogDist) / 70 ) );
		}
		
		else
		{
			waterVolColor = scattCoeff * fogColor.xyz / ( extCoeff * (1 + SinBoverSinA) );
			
			if (lava)
				waterVolColor = float3(1,0.33,0.1);
				

			waterVolColor *= ( exp( extCoeff * (1 + SinBoverSinA) * eyepos.z*depthDarkness / 70 ) - exp( -extCoeff * (1 + SinBoverSinA) * (waterfogDist) / 70 ) );
		}
		
		color.rgb *= exp( -extCoeff * ( max(-world_pos.z + eyepos.z*(1-depthDarkness*depthDarkness), 0) + waterfogDist ) / 70 );
		color.rgb += waterVolColor;
		float2 screenPos = IN.UVCoord / rcpres;
		color.rgb += ditherMat[ screenPos.x%4 ][ screenPos.y%4 ] / 255;
	}
	
    return float4(max(0,color.rgb),1);
}

technique t0
<
	int group = EFFECTGROUP_MAIN;
	int fxclass = EFFECTCLASS_WATER;
	int conditions = EFFECTCOND_HASWATER | EFFECTCOND_ZBUFFER;
>
{
	pass
	{
		VertexShader = compile vs_3_0 FrameVS();
		PixelShader = compile ps_3_0 Water();
	}
}